program start_VM;

uses Windows, SysUtils, Dialogs, Math, Classes, Controls, ShellApi, TNTSysUtils;

type
   _STORAGE_DEVICE_NUMBER = record
      DeviceType: DWORD;
      DeviceNumber: DWORD;
      PartitionNumber: DWORD;
   end;
   STORAGE_DEVICE_NUMBER = _STORAGE_DEVICE_NUMBER;
   TPWideCharArray = array[0..0] of PWideChar;

const
   METHOD_BUFFERED = 0;
   FILE_ANY_ACCESS = 0;
   FILE_DEVICE_FILE_SYSTEM = $00000009;
   FILE_DEVICE_MASS_STORAGE = $0000002D;
   IOCTL_STORAGE_BASE = FILE_DEVICE_MASS_STORAGE;
   IOCTL_STORAGE_GET_DEVICE_NUMBER = (
      (IOCTL_STORAGE_BASE shl 16) or (FILE_ANY_ACCESS shl 14) or
      ($0420 shl 2) or METHOD_BUFFERED);

var
   VolumesHandles: array of THandle;
   LastError: Integer;

procedure FlushToDisk(sDriveLetter: Char);
var
   hDrive: THandle;
begin
   hDrive := 0;
   try
      hDrive := CreateFile(PAnsiChar('\\.\' + sDriveLetter + ':'), GENERIC_READ or
         GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE, nil,
         OPEN_EXISTING, 0, 0);
      FlushFileBuffers(hDrive);
   finally
      CloseHandle(hDrive);
   end;
end;

function DismountVolume(ADrive: char): Boolean;
const
   FSCTL_DISMOUNT_VOLUME = (9 shl 16) or (0 shl 14) or (8 shl 2) or 0;
var
   VolumeName: string;
   BytesReturned: Cardinal;
begin
   Result := False;
   VolumeName := Format('\\.\%s:', [ADrive]);
   SetLength(VolumesHandles, Length(VolumesHandles) + 1);
   VolumesHandles[High(VolumesHandles)] := CreateFile(PChar(VolumeName), GENERIC_READ or GENERIC_WRITE,
      FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
   if VolumesHandles[High(VolumesHandles)] = INVALID_HANDLE_VALUE then
   begin
      LastError := GetLastError;
      Exit;
   end;
   Result := DeviceIoControl(VolumesHandles[High(VolumesHandles)], FSCTL_DISMOUNT_VOLUME, nil, 0,
      nil, 0, BytesReturned, nil);
   if not Result then
      LastError := GetLastError;
end;

function StrLCopyW(Dest, Source: PWideChar; MaxLen: Cardinal): PWideChar;
var
   Count: Cardinal;
begin
   Result := Dest;
   Count := 0;
   while (Count < MaxLen) and (Source^ <> #0) do
   begin
      Dest^ := Source^;
      Inc(Source);
      Inc(Dest);
      Inc(Count);
   end;
   Dest^ := #0;
end;

function StrPCopyW(Dest: PWideChar; const Source: WideString): PWideChar;
begin
   Result := StrLCopyW(Dest, PWideChar(Source), Length(Source));
end;

var
   zAppName: array[0..512] of WideChar;
   StartupInfo: TStartupInfo;
   ProcessInfo: TProcessInformation;
   Res: DWORD;
   i, udn, WaitTime: Integer;
   VolLet: array of Char;
   hVolume: THandle;
   dwBytesReturned: DWORD;
   sdn: STORAGE_DEVICE_NUMBER;
   ComLine, QFolder: WideString;
   NumArgs: Integer;
   TempArgs: PPWideChar;
   ParamStrW: array of WideString;
   ParamCountW: Integer;
   PQFolder: PWideChar;
   ErrorMode: Word;
begin

   SetLength(ParamStrW, 0);
   ParamCountW := -1;
   TempArgs := nil;
   try
      TempArgs := CommandLineToArgvW(GetCommandLineW, NumArgs);
   except
   end;
   if TempArgs <> nil then
   begin
      SetLength(ParamStrW, NumArgs);
      ParamCountW := NumArgs - 1;
      try
         for i := 0 to NumArgs - 1 do
            ParamStrW[i] := TPWideCharArray(TempArgs^)[i];
      except
      end;
      try
         LocalFree(THandle(TempArgs));
      except
      end;
   end;
   if ParamCount > ParamCountW then
   begin
      SetLength(ParamStrW, ParamCount + 1);
      for i := ParamCountW + 1 to ParamCount do
         ParamStrW[i] := ParamStr(i);
   end;

   if ParamCountW < 3 then
   begin
      MessageBox(0, 'Insufficient command line parameters !'#13#10'Please use "[USB_drive_number] [WaitTime] [QEMUexe] [QEMU_parameters]"', 'Warning', MB_OK or MB_ICONWARNING);
      Exit;
   end;

   udn := Min(Max(StrToIntDef(ParamStrW[1], 1), 0), 99);
   WaitTime := Min(Max(StrToIntDef(ParamStrW[2], 500), 1), 10000);

   ErrorMode := SetErrorMode(SEM_FailCriticalErrors);
   SetLength(VolLet, 0);
   for i := Byte('C') to Byte('Z') do
   begin
      case GetDriveType(PChar(Char(i) + ':\')) of
         DRIVE_REMOVABLE, DRIVE_FIXED:
            begin
               try
                  hVolume := CreateFile(PAnsiChar('\\.\' + Char(i) + ':'), 0, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, 0, 0);
               except
                  hVolume := INVALID_HANDLE_VALUE;
               end;
               if hVolume <> INVALID_HANDLE_VALUE then
               begin
                  dwBytesReturned := 0;
                  if DeviceIoControl(hVolume, IOCTL_STORAGE_GET_DEVICE_NUMBER, nil, 0, @sdn, sizeof(sdn), dwBytesReturned, nil) then
                     if sdn.DeviceNumber = Cardinal(udn) then
                     begin
                        SetLength(VolLet, Length(VolLet) + 1);
                        VolLet[High(VolLet)] := Char(i);
                     end;
               end;
            end
      else
         Continue;
      end;
   end;
   SetErrorMode(ErrorMode);

   SetLength(VolumesHandles, 0);

   for i := 0 to High(VolLet) do
   begin
      FlushToDisk(VolLet[i]);
      Sleep(WaitTime);
      if not DisMountVolume(VolLet[i]) then
         if MessageBox(0, PChar('Unable to dismount volume ''' + VolLet[i] + ''' !'#13#10'Possible reason: ' + SysErrorMessage(LastError) + #13#10'Are you sure you want to continue...?'), PChar('Warning'), MB_YESNO or MB_ICONWARNING or MB_DEFBUTTON2) = mrNO then
            Exit;
   end;

   ComLine := '';
   for i := 3 to ParamCountW do
      ComLine := ComLine + '"' + ParamStrW[i] + '" ';

   StrPCopyW(zAppName, ComLine);
   FillChar(StartupInfo, Sizeof(StartupInfo), #0);
   StartupInfo.cb := Sizeof(StartupInfo);
   StartupInfo.dwFlags := STARTF_USESHOWWINDOW;
   StartupInfo.wShowWindow := SW_MINIMIZE;
   QFolder := WideExtractFilePath(ParamStrW[3]);
   if QFolder <> '' then
      PQFolder := PWideChar(QFolder)
   else
      PQFolder := nil;
   if CreateProcessW(nil,
      zAppName,
      nil,
      nil,
      false,
      CREATE_NEW_CONSOLE or
      HIGH_PRIORITY_CLASS,
      nil,
      PQFolder,
      StartupInfo,
      ProcessInfo) then
   begin
      WaitforSingleObject(ProcessInfo.hProcess, INFINITE);
      GetExitCodeProcess(ProcessInfo.hProcess, Res);
      CloseHandle(ProcessInfo.hProcess);
      CloseHandle(ProcessInfo.hThread);
   end
   else
      ShowMessage('Unable to launch QEMU !');
   for i := 0 to High(VolumesHandles) do
      try
         CloseHandle(VolumesHandles[i]);
      except
      end;
end.

